@@ -55,7 +55,6 @@ func TestPartsRequired(t *testing.T) {
 }
 
 func TestCalculateEvenSplits(t *testing.T) {
-
 	testCases := []struct {
 		// input size and source object
 		size int64
@@ -69,47 +68,77 @@ func TestCalculateEvenSplits(t *testing.T) {
 		{1, CopySrcOptions{Start: 0}, []int64{0}, []int64{0}},
 
 		{gb1, CopySrcOptions{Start: -1}, []int64{0, 536870912}, []int64{536870911, 1073741823}},
-		{gb5, CopySrcOptions{Start: -1},
-			[]int64{0, 536870912, 1073741824, 1610612736, 2147483648, 2684354560,
-				3221225472, 3758096384, 4294967296, 4831838208},
-			[]int64{536870911, 1073741823, 1610612735, 2147483647, 2684354559, 3221225471,
-				3758096383, 4294967295, 4831838207, 5368709119},
+		{
+			gb5,
+			CopySrcOptions{Start: -1},
+			[]int64{
+				0, 536870912, 1073741824, 1610612736, 2147483648, 2684354560,
+				3221225472, 3758096384, 4294967296, 4831838208,
+			},
+			[]int64{
+				536870911, 1073741823, 1610612735, 2147483647, 2684354559, 3221225471,
+				3758096383, 4294967295, 4831838207, 5368709119,
+			},
 		},
 
 		// 2 part splits
-		{gb5p1, CopySrcOptions{Start: -1},
-			[]int64{0, 536870913, 1073741825, 1610612737, 2147483649, 2684354561,
-				3221225473, 3758096385, 4294967297, 4831838209},
-			[]int64{536870912, 1073741824, 1610612736, 2147483648, 2684354560, 3221225472,
-				3758096384, 4294967296, 4831838208, 5368709120},
+		{
+			gb5p1,
+			CopySrcOptions{Start: -1},
+			[]int64{
+				0, 536870913, 1073741825, 1610612737, 2147483649, 2684354561,
+				3221225473, 3758096385, 4294967297, 4831838209,
+			},
+			[]int64{
+				536870912, 1073741824, 1610612736, 2147483648, 2684354560, 3221225472,
+				3758096384, 4294967296, 4831838208, 5368709120,
+			},
 		},
-		{gb5p1, CopySrcOptions{Start: -1},
-			[]int64{0, 536870913, 1073741825, 1610612737, 2147483649, 2684354561,
-				3221225473, 3758096385, 4294967297, 4831838209},
-			[]int64{536870912, 1073741824, 1610612736, 2147483648, 2684354560, 3221225472,
-				3758096384, 4294967296, 4831838208, 5368709120},
+		{
+			gb5p1,
+			CopySrcOptions{Start: -1},
+			[]int64{
+				0, 536870913, 1073741825, 1610612737, 2147483649, 2684354561,
+				3221225473, 3758096385, 4294967297, 4831838209,
+			},
+			[]int64{
+				536870912, 1073741824, 1610612736, 2147483648, 2684354560, 3221225472,
+				3758096384, 4294967296, 4831838208, 5368709120,
+			},
 		},
 
 		// 3 part splits
-		{gb10p1, CopySrcOptions{Start: -1},
-			[]int64{0, 536870913, 1073741825, 1610612737, 2147483649, 2684354561,
+		{
+			gb10p1,
+			CopySrcOptions{Start: -1},
+			[]int64{
+				0, 536870913, 1073741825, 1610612737, 2147483649, 2684354561,
 				3221225473, 3758096385, 4294967297, 4831838209, 5368709121,
 				5905580033, 6442450945, 6979321857, 7516192769, 8053063681,
-				8589934593, 9126805505, 9663676417, 10200547329},
-			[]int64{536870912, 1073741824, 1610612736, 2147483648, 2684354560,
+				8589934593, 9126805505, 9663676417, 10200547329,
+			},
+			[]int64{
+				536870912, 1073741824, 1610612736, 2147483648, 2684354560,
 				3221225472, 3758096384, 4294967296, 4831838208, 5368709120,
 				5905580032, 6442450944, 6979321856, 7516192768, 8053063680,
-				8589934592, 9126805504, 9663676416, 10200547328, 10737418240},
+				8589934592, 9126805504, 9663676416, 10200547328, 10737418240,
+			},
 		},
-		{gb10p2, CopySrcOptions{Start: -1},
-			[]int64{0, 536870913, 1073741826, 1610612738, 2147483650, 2684354562,
+		{
+			gb10p2,
+			CopySrcOptions{Start: -1},
+			[]int64{
+				0, 536870913, 1073741826, 1610612738, 2147483650, 2684354562,
 				3221225474, 3758096386, 4294967298, 4831838210, 5368709122,
 				5905580034, 6442450946, 6979321858, 7516192770, 8053063682,
-				8589934594, 9126805506, 9663676418, 10200547330},
-			[]int64{536870912, 1073741825, 1610612737, 2147483649, 2684354561,
+				8589934594, 9126805506, 9663676418, 10200547330,
+			},
+			[]int64{
+				536870912, 1073741825, 1610612737, 2147483649, 2684354561,
 				3221225473, 3758096385, 4294967297, 4831838209, 5368709121,
 				5905580033, 6442450945, 6979321857, 7516192769, 8053063681,
-				8589934593, 9126805505, 9663676417, 10200547329, 10737418241},
+				8589934593, 9126805505, 9663676417, 10200547329, 10737418241,
+			},
 		},
 	}
 
