@@ -19,8 +19,9 @@ package minio
 
 import (
 	"bytes"
+	"context"
 	"encoding/xml"
-	"io/ioutil"
+	"io"
 	"net/http"
 	"net/url"
 	"path"
@@ -125,7 +126,6 @@ func TestGetBucketLocationRequest(t *testing.T) {
 		}
 
 		return req, nil
-
 	}
 	// Info for 'Client' creation.
 	// Will be used as arguments for 'NewClient'.
@@ -236,7 +236,7 @@ func TestGetBucketLocationRequest(t *testing.T) {
 			}
 		}
 
-		actualReq, err := client.getBucketLocationRequest(testCase.bucketName)
+		actualReq, err := client.getBucketLocationRequest(context.Background(), testCase.bucketName)
 		if err != nil && testCase.shouldPass {
 			t.Errorf("Test %d: Expected to pass, but failed with: <ERROR> %s", i+1, err.Error())
 		}
@@ -279,7 +279,7 @@ func TestGetBucketLocationRequest(t *testing.T) {
 // generates http response with bucket location set in the body.
 func generateLocationResponse(resp *http.Response, bodyContent []byte) (*http.Response, error) {
 	resp.StatusCode = http.StatusOK
-	resp.Body = ioutil.NopCloser(bytes.NewBuffer(bodyContent))
+	resp.Body = io.NopCloser(bytes.NewBuffer(bodyContent))
 	return resp, nil
 }
 
