@@ -20,16 +20,16 @@ package minio
 import (
 	"bytes"
 	"context"
-	"log"
+	"io"
+	"math/rand"
 	"net/http"
 	"os"
 	"strconv"
 	"testing"
 	"time"
 
-	"math/rand"
-
 	"github.com/minio/minio-go/v7/pkg/credentials"
+	"github.com/minio/minio-go/v7/pkg/encrypt"
 )
 
 const (
@@ -39,31 +39,6 @@ const (
 	enableSecurity = "ENABLE_HTTPS"
 )
 
-const letterBytes = "abcdefghijklmnopqrstuvwxyz01234569"
-const (
-	letterIdxBits = 6                    // 6 bits to represent a letter index
-	letterIdxMask = 1<<letterIdxBits - 1 // All 1-bits, as many as letterIdxBits
-	letterIdxMax  = 63 / letterIdxBits   // # of letter indices fitting in 63 bits
-)
-
-// randString generates random names and prepends them with a known prefix.
-func randString(n int, src rand.Source, prefix string) string {
-	b := make([]byte, n)
-	// A rand.Int63() generates 63 random bits, enough for letterIdxMax letters!
-	for i, cache, remain := n-1, src.Int63(), letterIdxMax; i >= 0; {
-		if remain == 0 {
-			cache, remain = src.Int63(), letterIdxMax
-		}
-		if idx := int(cache & letterIdxMask); idx < len(letterBytes) {
-			b[i] = letterBytes[idx]
-			i--
-		}
-		cache >>= letterIdxBits
-		remain--
-	}
-	return prefix + string(b[0:30-len(prefix)])
-}
-
 // Tests for Core GetObject() function.
 func TestGetObjectCore(t *testing.T) {
 	if testing.Short() {
@@ -462,7 +437,7 @@ func TestCoreCopyObject(t *testing.T) {
 	cuploadInfo, err := c.CopyObject(context.Background(), bucketName, objectName, destBucketName, destObjectName, map[string]string{
 		"X-Amz-Metadata-Directive": "REPLACE",
 		"Content-Type":             "application/javascript",
-	})
+	}, CopySrcOptions{}, PutObjectOptions{})
 	if err != nil {
 		t.Fatal("Error:", err, bucketName, objectName, destBucketName, destObjectName)
 	}
@@ -610,7 +585,7 @@ func TestCoreCopyObjectPart(t *testing.T) {
 	}
 
 	// Complete the multipart upload
-	_, err = c.CompleteMultipartUpload(context.Background(), destBucketName, destObjectName, uploadID, []CompletePart{fstPart, sndPart, lstPart})
+	_, err = c.CompleteMultipartUpload(context.Background(), destBucketName, destObjectName, uploadID, []CompletePart{fstPart, sndPart, lstPart}, PutObjectOptions{})
 	if err != nil {
 		t.Fatal("Error:", err, destBucketName, destObjectName)
 	}
@@ -802,17 +777,17 @@ func TestCoreGetObjectMetadata(t *testing.T) {
 	_, err = core.PutObject(context.Background(), bucketName, "my-objectname",
 		bytes.NewReader([]byte("hello")), 5, "", "", putopts)
 	if err != nil {
-		log.Fatalln(err)
+		t.Fatal(err)
 	}
 
 	reader, objInfo, _, err := core.GetObject(context.Background(), bucketName, "my-objectname", GetObjectOptions{})
 	if err != nil {
-		log.Fatalln(err)
+		t.Fatal(err)
 	}
-	defer reader.Close()
+	reader.Close()
 
 	if objInfo.Metadata.Get("X-Amz-Meta-Key-1") != "Val-1" {
-		log.Fatalln("Expected metadata to be available but wasn't")
+		t.Fatal("Expected metadata to be available but wasn't")
 	}
 
 	err = core.RemoveObject(context.Background(), bucketName, "my-objectname", RemoveObjectOptions{})
@@ -824,3 +799,85 @@ func TestCoreGetObjectMetadata(t *testing.T) {
 		t.Fatal("Error:", err)
 	}
 }
+
+func TestCoreMultipartUpload(t *testing.T) {
+	if testing.Short() {
+		t.Skip("skipping functional tests for the short runs")
+	}
+
+	// Instantiate new minio client object.
+	core, err := NewCore(
+		os.Getenv(serverEndpoint),
+		&Options{
+			Creds:  credentials.NewStaticV4(os.Getenv(accessKey), os.Getenv(secretKey), ""),
+			Secure: mustParseBool(os.Getenv(enableSecurity)),
+		})
+	if err != nil {
+		t.Fatal("Error:", err)
+	}
+
+	bucketName := randString(60, rand.NewSource(time.Now().UnixNano()), "minio-go-test")
+	// Make a new bucket.
+	err = core.MakeBucket(context.Background(), bucketName, MakeBucketOptions{Region: "us-east-1"})
+	if err != nil {
+		t.Fatal("Error:", err, bucketName)
+	}
+	objectName := randString(60, rand.NewSource(time.Now().UnixNano()), "")
+
+	objectContentType := "binary/octet-stream"
+	metadata := make(map[string]string)
+	metadata["Content-Type"] = objectContentType
+	putopts := PutObjectOptions{
+		UserMetadata: metadata,
+	}
+	uploadID, err := core.NewMultipartUpload(context.Background(), bucketName, objectName, putopts)
+	if err != nil {
+		t.Fatal("Error:", err, bucketName, objectName)
+	}
+	buf := bytes.Repeat([]byte("a"), 32*1024*1024)
+	r := bytes.NewReader(buf)
+	partBuf := make([]byte, 100*1024*1024)
+	parts := make([]CompletePart, 0, 5)
+	partID := 0
+	for {
+		n, err := r.Read(partBuf)
+		if err != nil && err != io.EOF {
+			t.Fatal("Error:", err)
+		}
+		if err == io.EOF {
+			break
+		}
+		if n > 0 {
+			partID++
+			data := bytes.NewReader(partBuf[:n])
+			dataLen := int64(len(partBuf[:n]))
+			objectPart, err := core.PutObjectPart(context.Background(), bucketName, objectName, uploadID, partID, data, dataLen, "", "", encrypt.NewSSE())
+			if err != nil {
+				t.Fatal("Error:", err, bucketName, objectName)
+			}
+			parts = append(parts, CompletePart{
+				PartNumber: partID,
+				ETag:       objectPart.ETag,
+			})
+		}
+	}
+	objectParts, err := core.listObjectParts(context.Background(), bucketName, objectName, uploadID)
+	if err != nil {
+		t.Fatal("Error:", err)
+	}
+	if len(objectParts) != len(parts) {
+		t.Fatal("Error", len(objectParts), len(parts))
+	}
+	_, err = core.CompleteMultipartUpload(context.Background(), bucketName, objectName, uploadID, parts, putopts)
+	if err != nil {
+		t.Fatal("Error:", err)
+	}
+
+	if err := core.RemoveObject(context.Background(), bucketName, objectName, RemoveObjectOptions{}); err != nil {
+		t.Fatal("Error: ", err)
+	}
+
+	if err := core.RemoveBucket(context.Background(), bucketName); err != nil {
+		t.Fatal("Error: ", err)
+	}
+}
