@@ -1,6 +1,6 @@
 /*
- * Minio Go Library for Amazon S3 Compatible Cloud Storage
- * Copyright 2015-2017 Minio, Inc.
+ * MinIO Go Library for Amazon S3 Compatible Cloud Storage
+ * Copyright 2015-2017 MinIO, Inc.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -17,7 +17,7 @@
 
 package policy
 
-import "github.com/minio/minio-go/pkg/set"
+import "github.com/minio/minio-go/v7/pkg/set"
 
 // ConditionKeyMap - map of policy condition key and value.
 type ConditionKeyMap map[string]set.StringSet
@@ -46,9 +46,7 @@ func (ckm ConditionKeyMap) Remove(key string, value set.StringSet) {
 
 // RemoveKey - removes key and its value.
 func (ckm ConditionKeyMap) RemoveKey(key string) {
-	if _, ok := ckm[key]; ok {
-		delete(ckm, key)
-	}
+	delete(ckm, key)
 }
 
 // CopyConditionKeyMap - returns new copy of given ConditionKeyMap.
@@ -91,9 +89,7 @@ func (cond ConditionMap) Add(condKey string, condKeyMap ConditionKeyMap) {
 
 // Remove - removes condition key and its value.
 func (cond ConditionMap) Remove(condKey string) {
-	if _, ok := cond[condKey]; ok {
-		delete(cond, condKey)
-	}
+	delete(cond, condKey)
 }
 
 // mergeConditionMap - returns new ConditionMap which contains merged key/value of two ConditionMap.
