@@ -1,6 +1,6 @@
 /*
- * Minio Go Library for Amazon S3 Compatible Cloud Storage
- * Copyright 2015-2017 Minio, Inc.
+ * MinIO Go Library for Amazon S3 Compatible Cloud Storage
+ * Copyright 2015-2017 MinIO, Inc.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -20,6 +20,7 @@ package s3utils
 import (
 	"errors"
 	"net/url"
+	"reflect"
 	"testing"
 )
 
@@ -37,14 +38,26 @@ func TestGetRegionFromURL(t *testing.T) {
 			u:              url.URL{Host: "s3.cn-north-1.amazonaws.com.cn"},
 			expectedRegion: "cn-north-1",
 		},
+		{
+			u:              url.URL{Host: "s3.dualstack.cn-north-1.amazonaws.com.cn"},
+			expectedRegion: "cn-north-1",
+		},
 		{
 			u:              url.URL{Host: "s3.cn-northwest-1.amazonaws.com.cn"},
 			expectedRegion: "cn-northwest-1",
 		},
+		{
+			u:              url.URL{Host: "s3.dualstack.cn-northwest-1.amazonaws.com.cn"},
+			expectedRegion: "cn-northwest-1",
+		},
 		{
 			u:              url.URL{Host: "s3-fips-us-gov-west-1.amazonaws.com"},
 			expectedRegion: "us-gov-west-1",
 		},
+		{
+			u:              url.URL{Host: "s3-fips.us-gov-west-1.amazonaws.com"},
+			expectedRegion: "us-gov-west-1",
+		},
 		{
 			u:              url.URL{Host: "s3-us-gov-west-1.amazonaws.com"},
 			expectedRegion: "us-gov-west-1",
@@ -73,6 +86,47 @@ func TestGetRegionFromURL(t *testing.T) {
 			u:              url.URL{Host: "s3-external-1.amazonaws.com"},
 			expectedRegion: "",
 		},
+		{
+			u: url.URL{
+				Host: "s3.kubernetesfrontendlb-caf78da2b1f7516c.elb.us-west-2.amazonaws.com",
+			},
+			expectedRegion: "",
+		},
+		{
+			u: url.URL{
+				Host: "s3.kubernetesfrontendlb-caf78da2b1f7516c.elb.amazonaws.com",
+			},
+			expectedRegion: "",
+		},
+		{
+			u: url.URL{
+				Host: "s3.kubernetesfrontendlb-caf78da2b1f7516c.elb.amazonaws.com.cn",
+			},
+		},
+		{
+			u: url.URL{
+				Host: "bucket.vpce-1a2b3c4d-5e6f.s3.us-east-1.vpce.amazonaws.com",
+			},
+			expectedRegion: "us-east-1",
+		},
+		{
+			u: url.URL{
+				Host: "accesspoint.vpce-1a2b3c4d-5e6f.s3.us-east-1.vpce.amazonaws.com",
+			},
+			expectedRegion: "us-east-1",
+		},
+		{
+			u: url.URL{
+				Host: "s3-fips.us-east-1.amazonaws.com",
+			},
+			expectedRegion: "us-east-1",
+		},
+		{
+			u: url.URL{
+				Host: "s3-fips.dualstack.us-west-1.amazonaws.com",
+			},
+			expectedRegion: "us-west-1",
+		},
 	}
 
 	for i, testCase := range testCases {
@@ -99,6 +153,7 @@ func TestIsValidDomain(t *testing.T) {
 		{"s3.amz.test.com", true},
 		{"s3.%%", false},
 		{"localhost", true},
+		{"localhost.", true}, // http://www.dns-sd.org/trailingdotsindomainnames.html
 		{"-localhost", false},
 		{"", false},
 		{"\n \t", false},
@@ -134,7 +189,6 @@ func TestIsValidIP(t *testing.T) {
 			t.Errorf("Test %d: Expected isValidIP to be '%v' for input \"%s\", but found it to be '%v' instead", i+1, testCase.result, testCase.ip, result)
 		}
 	}
-
 }
 
 // Tests validate virtual host validator.
@@ -192,6 +246,8 @@ func TestIsAmazonEndpoint(t *testing.T) {
 		{"https://s3-us-west-1.amazonaws.com", true},
 		{"https://s3.us-west-1.amazonaws.com", true},
 		{"https://s3.dualstack.us-west-1.amazonaws.com", true},
+		{"https://bucket.vpce-1a2b3c4d-5e6f.s3.us-east-1.vpce.amazonaws.com", true},
+		{"https://accesspoint.vpce-1a2b3c4d-5e6f.s3.us-east-1.vpce.amazonaws.com", true},
 	}
 
 	for i, testCase := range testCases {
@@ -204,7 +260,6 @@ func TestIsAmazonEndpoint(t *testing.T) {
 			t.Errorf("Test %d: Expected isAmazonEndpoint to be '%v' for input \"%s\", but found it to be '%v' instead", i+1, testCase.result, testCase.url, result)
 		}
 	}
-
 }
 
 // Tests validate Google Cloud end point validator.
@@ -237,7 +292,6 @@ func TestIsGoogleEndpoint(t *testing.T) {
 			t.Errorf("Test %d: Expected isGoogleEndpoint to be '%v' for input \"%s\", but found it to be '%v' instead", i+1, testCase.result, testCase.url, result)
 		}
 	}
-
 }
 
 func TestPercentEncodeSlash(t *testing.T) {
@@ -260,7 +314,6 @@ func TestPercentEncodeSlash(t *testing.T) {
 				i+1, testCase.input, testCase.output,
 				receivedOutput,
 			)
-
 		}
 	}
 }
@@ -294,6 +347,81 @@ func TestQueryEncode(t *testing.T) {
 	}
 }
 
+// Tests tag decode to map
+func TestTagDecode(t *testing.T) {
+	testCases := []struct {
+		// canonical input
+		canonicalInput string
+
+		// Expected result.
+		resultMap map[string]string
+	}{
+		{"k=thisisthe%25url", map[string]string{"k": "thisisthe%url"}},
+		{"k=%E6%9C%AC%E8%AA%9E", map[string]string{"k": "本語"}},
+		{"k=%E6%9C%AC%E8%AA%9E.1", map[string]string{"k": "本語.1"}},
+		{"k=%3E123", map[string]string{"k": ">123"}},
+		{"k=myurl%23link", map[string]string{"k": "myurl#link"}},
+		{"k=space%20in%20url", map[string]string{"k": "space in url"}},
+		{"k=url%2Bpath", map[string]string{"k": "url+path"}},
+		{"k=url%2Fpath", map[string]string{"k": "url/path"}},
+	}
+
+	for _, testCase := range testCases {
+		testCase := testCase
+		t.Run("", func(t *testing.T) {
+			gotResult := TagDecode(testCase.canonicalInput)
+			if !reflect.DeepEqual(testCase.resultMap, gotResult) {
+				t.Errorf("Expected %s, got %s", testCase.resultMap, gotResult)
+			}
+		})
+	}
+}
+
+// Tests tag encode function for user tags.
+func TestTagEncode(t *testing.T) {
+	testCases := []struct {
+		// Input.
+		inputMap map[string]string
+		// Expected result.
+		result string
+	}{
+		{map[string]string{
+			"k": "thisisthe%url",
+		}, "k=thisisthe%25url"},
+		{map[string]string{
+			"k": "本語",
+		}, "k=%E6%9C%AC%E8%AA%9E"},
+		{map[string]string{
+			"k": "本語.1",
+		}, "k=%E6%9C%AC%E8%AA%9E.1"},
+		{map[string]string{
+			"k": ">123",
+		}, "k=%3E123"},
+		{map[string]string{
+			"k": "myurl#link",
+		}, "k=myurl%23link"},
+		{map[string]string{
+			"k": "space in url",
+		}, "k=space%20in%20url"},
+		{map[string]string{
+			"k": "url+path",
+		}, "k=url%2Bpath"},
+		{map[string]string{
+			"k": "url/path",
+		}, "k=url%2Fpath"},
+	}
+
+	for _, testCase := range testCases {
+		testCase := testCase
+		t.Run("", func(t *testing.T) {
+			gotResult := TagEncode(testCase.inputMap)
+			if testCase.result != gotResult {
+				t.Errorf("Expected %s, got %s", testCase.result, gotResult)
+			}
+		})
+	}
+}
+
 // Tests validate the URL path encoder.
 func TestEncodePath(t *testing.T) {
 	testCases := []struct {
@@ -309,6 +437,7 @@ func TestEncodePath(t *testing.T) {
 		{"myurl#link", "myurl%23link"},
 		{"space in url", "space%20in%20url"},
 		{"url+path", "url%2Bpath"},
+		{"url/path", "url/path"},
 	}
 
 	for i, testCase := range testCases {
@@ -332,9 +461,11 @@ func TestIsValidBucketName(t *testing.T) {
 		{".mybucket", errors.New("Bucket name contains invalid characters"), false},
 		{"$mybucket", errors.New("Bucket name contains invalid characters"), false},
 		{"mybucket-", errors.New("Bucket name contains invalid characters"), false},
-		{"my", errors.New("Bucket name cannot be smaller than 3 characters"), false},
+		{"my", errors.New("Bucket name cannot be shorter than 3 characters"), false},
 		{"", errors.New("Bucket name cannot be empty"), false},
 		{"my..bucket", errors.New("Bucket name contains invalid characters"), false},
+		{"my.-bucket", errors.New("Bucket name contains invalid characters"), false},
+		{"my-.bucket", errors.New("Bucket name contains invalid characters"), false},
 		{"192.168.1.168", errors.New("Bucket name cannot be an ip address"), false},
 		{":bucketname", errors.New("Bucket name contains invalid characters"), false},
 		{"_bucketName", errors.New("Bucket name contains invalid characters"), false},
@@ -362,7 +493,6 @@ func TestIsValidBucketName(t *testing.T) {
 		}
 
 	}
-
 }
 
 // Tests validate the bucket name validator stricter.
@@ -378,9 +508,11 @@ func TestIsValidBucketNameStrict(t *testing.T) {
 		{".mybucket", errors.New("Bucket name contains invalid characters"), false},
 		{"$mybucket", errors.New("Bucket name contains invalid characters"), false},
 		{"mybucket-", errors.New("Bucket name contains invalid characters"), false},
-		{"my", errors.New("Bucket name cannot be smaller than 3 characters"), false},
+		{"my", errors.New("Bucket name cannot be shorter than 3 characters"), false},
 		{"", errors.New("Bucket name cannot be empty"), false},
 		{"my..bucket", errors.New("Bucket name contains invalid characters"), false},
+		{"my.-bucket", errors.New("Bucket name contains invalid characters"), false},
+		{"my-.bucket", errors.New("Bucket name contains invalid characters"), false},
 		{"192.168.1.168", errors.New("Bucket name cannot be an ip address"), false},
 		{"Mybucket", errors.New("Bucket name contains invalid characters"), false},
 		{"my.bucket.com", nil, true},
@@ -404,5 +536,4 @@ func TestIsValidBucketNameStrict(t *testing.T) {
 		}
 
 	}
-
 }
